/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "tstore.h"
#include "test.h"

uint64_t store_size;
int store_fd;
char *namebuf;
char *storename;
int maxused=0,maxsecused=0,overflows=0;

int zblock(uint32_t *buf)
{
  int i;
  for (i=0;i<128;++i)
    if (buf[i]) return 0;
  return 1;
}

FILE *hash_file;
uint64_t filesize=0;
wtree_t hsec_tree=NULL;

wtree_t allocated_node=NULL;

wtree_t get_allocated_node()
{
  if (!allocated_node) allocated_node=allocnode();
  return allocated_node;
}

wtree_t put_allocated_node()
{
  wtree_t res=insert(&hsec_tree,allocated_node);
  if (!res) {
    res=allocated_node;
    allocated_node=NULL;
  }
  return res;
}

#define MAX_BATCH 250000
struct si{
  uint64_t *l;
  union {
    uint8_t *d;
    uint64_t l;
  } u;
} sec_info[MAX_BATCH+2];

int load_sector(int fd, struct si *si)
{
  wtree_t node=get_allocated_node();
  int r,z;
  r=read_and_hash(fd,datap(node));
  if((r<1)||(r>512)) return r;
  z=zblock((void*)(datap(node)+16));
  node=put_allocated_node();
  if(z) {
    si->l=&(si->u.l);
    si->u.l=0;
  } else {
    si->l=locationp(node);
    si->u.d=datap(node);
  }
  return r;
}

struct block_num_queue {
  struct block_num_queue *next,*prev;
  int count;
  uint64_t numbers[102];
} *queue_head;

void put_block_number(struct block_num_queue *head,
		      uint64_t blocknumber)
{
  if ((head->prev == head)||(head->prev->count==102)) {
    struct block_num_queue *new;
    NEW(new);
    new->prev=head->prev;
    new->next=head;
    new->count=0;
    head->prev->next=new;
    head->prev=new;
  }
  head->prev->numbers[head->prev->count++]=blocknumber;
}

void init_queue()
{
  NEW(queue_head);
  queue_head->next=queue_head->prev=queue_head;
  queue_head->count=-1;
}

void batch2secnum()
{
  uint64_t i;

  assert(fseek(hash_file,16,SEEK_SET)==0);

  for (i=1;i<store_size;++i) {
    uint8_t hash_sector_buf[HBS];
    wtree_t t;
    uint32_t hash_value;
    
    if (fread(&hash_value,1,4,hash_file)!=4) {
      perror(namebuf);
      exit(EXIT_FAILURE);
    }

    t=find_hash(hsec_tree,hash_value);
    if (t) {
      /* FIXME: Use pread instead */
      lseek(store_fd,512*i,SEEK_SET);
      switch(read_and_hash(store_fd,hash_sector_buf)) {
      case -1:
	perror("read");
	exit(EXIT_FAILURE);
      case 512:
	break;
      case 0:
	fprintf(stderr,"Short read\n");
	exit(EXIT_FAILURE);
      }
      t=find(t,hash_sector_buf);
      if (t) (*locationp(t))=i;
    }
  }
}

void secnum2queue(struct block_num_queue *q, uint64_t batchsize)
{
  uint64_t i;
  assert(fseek(hash_file,(store_size+3)*4,SEEK_SET)==0);
  lseek(store_fd,store_size*512,SEEK_SET);
  for (i=0;i<batchsize;++i) {
    uint64_t *l=sec_info[i].l;
    if (l != &(sec_info[i].u.l)) {
      if (!*l) {
	uint8_t *dp=sec_info[i].u.d;
	assert(write(store_fd,dp+16,512)==512);
	assert(fwrite(dp,1,4,hash_file)==4);
	*l=store_size++;
      }
    }
    put_block_number(q,*l);
  }
}

#undef DEBUG_OUT
void dumpsec(int sector_num)
{
#ifdef DEBUG_OUT
  int i;
  for (i=0;i<sector_num;++i) {
    /*
    wtree_t t=secp[i];
    if (t)
      printf("%d ",*locationp(t));
    else
      printf("_ ");
    */
    printf("%lld ",*sec_info[i].l);
  }
  printf("\n");
#endif
}

void dumpqn(struct block_num_queue *qn)
{
#ifdef DEBUG_OUT
  int i;
  printf("[ ");
  for (i=0;i<qn->count;++i)
    printf("%lld ",qn->numbers[i]);
  printf("]");
#endif
}

void dumpq(struct block_num_queue *q)
{
#ifdef DEBUG_OUT
  struct block_num_queue *p;
  for (p=q->next;p!=q;p=p->next)
    dumpqn(p);
  printf("\n");
#endif
}

int main(int argc, char ** argv)
{
  char formbuf[42];

  char *progbuf=getenv("TSTORE");
  if (progbuf) sprintf(formbuf,"%%%dlld",strlen(progbuf));

  init_queue();

  storename=argv[1];

  if ((argc != 3)&&(argc != 4)) {
    fprintf(stderr,"Uage: %s <store> <file> [timestamp]\n",argv[0]);
    return 1;
  }

  if (strlen(argv[2])>489) {
    fprintf(stderr,"%s: Name too long\n",argv[2]);
    return 1;
  }

  store_fd=open(argv[1],O_RDWR);
  if (store_fd==-1) {
    perror(argv[1]);
    return 1;
  }

  store_size=test_store_size(store_fd,1);
  store_error(store_size,argv[1]);
  lseek(store_fd,0,SEEK_END);

  namebuf=malloc(strlen(argv[1])+10);
  if (!namebuf) {
    perror("Out of memory");
    return 1;
  }

  /* Old hash stuff */
#if 0

  sprintf(namebuf,"%s.idx",argv[1]);

  hash_fd=open(namebuf,O_RDWR,0666);
  if (hash_fd==-1) {
    perror(namebuf);
    return 1;
  }

  buckets=size_in_blocks(hash_fd);
  if (!buckets) {
    fprintf(stderr,"Invalid hash index (rebuild?)\n");
    return 1;
  }

  hash_map=mmap(NULL,buckets*512,PROT_READ|PROT_WRITE,MAP_SHARED,hash_fd,0);
  if (hash_map==MAP_FAILED) {
    perror("mmap");
    return 1;
  }

  init_hash();
#endif

  /* New hash stuff */

  sprintf(namebuf,"%s.32h",argv[1]);

  hash_file=fopen(namebuf,"r+");
  if (!hash_file) {
    perror(namebuf);
    return 1;
  }

  if (fread(hash_key,1,16,hash_file)!=16) {
    perror(namebuf);
    return EXIT_FAILURE;
  }

  while(1) {

    int r;
    int batch_size=0;
    
    assert(hsec_tree == NULL);

    while((batch_size < MAX_BATCH)&&(r=load_sector(0,sec_info+batch_size))) {
      if ((r>512)||(r<0)) {
	perror("read");
	return 1;
      }
      filesize += r;
      if (progbuf) sprintf(progbuf,formbuf,filesize);
      assert(batch_size++ < MAX_BATCH);
    }

    TEST(batch_size);
    if (!batch_size) break;

    dumpsec(batch_size);
    batch2secnum();
    dumpsec(batch_size);
    secnum2queue(queue_head,batch_size);
    dumpsec(batch_size);

    discard(hsec_tree);
    hsec_tree = NULL;

    dumpq(queue_head);

  }

  if (filesize == 0) return 0;

  dumpq(queue_head);

  while (queue_head->next->count != 1) {
    struct block_num_queue *q=queue_head;
    init_queue();
    int batch_size=0;

    for (batch_size=0; q->next != q; ++batch_size) {
      int count=q->next->count;
      uint64_t *numbers=q->next->numbers;
      int z=1,s=0,i;
      assert(count>0);
      for (i=0;i<count;++i) {
	if (numbers[i]) {
	  z=0;
	  break;
	}
      }
      if (z||(numbers[0]>=0xC000000000ULL)) {
	s=0;
      } else {
	s=1;
	for(i=1;i<count;++i) {
	  if (numbers[i]!=numbers[i-1]+1) {
	    s=0;
	    break;
	  }
	}
      }
      if (z||s) {
	sec_info[batch_size].l=&(sec_info[batch_size].u.l);
	sec_info[batch_size].u.l=z?0ULL:numbers[0]+0x4000000000ULL;
      } else {
	wtree_t node=get_allocated_node();
	uint8_t *buf=datap(node);
	for (i=0;i<count;++i) PUT_BNUM(buf+(16+5*i),numbers[i]);
	hash_buf(buf,count*5);
	node=put_allocated_node();
	sec_info[batch_size].u.d=datap(node);
	sec_info[batch_size].l=locationp(node);
      }
      q->next=q->next->next;
    }

    TEST(batch_size);
    assert(batch_size);

    dumpsec(batch_size);
    batch2secnum();
    dumpsec(batch_size);
    secnum2queue(queue_head,batch_size);
    dumpsec(batch_size);

    discard(hsec_tree);
    hsec_tree = NULL;

    dumpq(queue_head);

  }

  /* Write the file header record */
  {
    wtree_t node=get_allocated_node();
    uint8_t *buf=datap(node)+16;
  
    memset(buf,0,512);
    assert(pread(store_fd,buf,5,13)==5);
    PUT_BNUM(buf+5,queue_head->next->numbers[0]);
    PUT_SIZE(buf+10,filesize);
    strcpy(buf+17,argv[2]);
    PUT_BNUM(buf+507,(argc==4)?atoi(argv[3]):time(NULL));

    hash_buf(datap(node),512);
    node=put_allocated_node();
    sec_info[0].u.d=datap(node);
    sec_info[0].l=locationp(node);

    batch2secnum();
    secnum2queue(queue_head,1);

    assert(queue_head->next->count==2);
    
    PUT_BNUM(buf,store_size);
    PUT_BNUM(buf+5,queue_head->next->numbers[1]);
    assert(pwrite(store_fd,buf,10,8)==10);

    return 0;
  }
}
