/*
    tstore - an archive format for archiving tar archives
    Copyright (C) 2004  Kasper Dupont

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    with permission from the author any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    Parts of this code may originate from other
    filesystem drivers also with GPL licenses:
    swapfs, minix, romfs, affs.
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/init.h>
#include <linux/string.h>
//#include <linux/locks.h>
#include <linux/buffer_head.h>
#include <linux/vfs.h>

//#include <asm/uaccess.h>

/* some random number */
#define TSTORE_MAGIC	0x74617273

#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/fs.h>
//#include <linux/locks.h>
#include <linux/init.h>
#include <linux/smp_lock.h>

#include <asm/uaccess.h>

#if 0
#define TSTORE_DBG(x) printk(KERN_DEBUG "TSTORE(%s): %s\n",__FUNCTION__,x)
#else
#define TSTORE_DBG(x) do { } while(0)
#endif
#define ENTER() TSTORE_DBG("ENTER")
#define TEST(__x) printk(KERN_DEBUG "TSTORE-TEST: %s = %x\n",#__x,(int)(__x))

static struct super_operations tstore_ops;

struct tstore_super {
  sector_t numblocks,lastentry;
};

uint64_t get_nbyteint(uint8_t *p, int n)
{
	uint64_t r=0;
	int i;
	for (i=0;i<n;++i)
		r=(r<<8)+p[i];
	return r;
}

#define GET_BNUM(_p) get_nbyteint(_p,5)
#define GET_SIZE(_p) get_nbyteint(_p,7)

int filelevels(uint64_t size)
{
	int levels=0;
	uint64_t maxsize=512;
	while(size>maxsize) {
		levels++;
		maxsize*=102;
	}
	return levels;
}

static int tstore_fill_super(struct super_block *s, void *data, int silent)
{
	struct tstore_super *ts_sb;
	struct buffer_head *bh;

	ENTER();

	if (!s) {
		printk(KERN_ERR "TSTORE: No super block! HUH?\n");
		return -EINVAL;
	}

	s->s_fs_info=ts_sb = kmalloc(sizeof(struct tstore_super), GFP_KERNEL);
	if (!ts_sb)
		return -ENOMEM;

	/* I would parse the options here, but there are none.. :) */

	if (!sb_set_blocksize(s, 512))
		goto outnobh;
	s->s_maxbytes = (1ULL)<<56;

	bh = sb_bread(s, 0);
	if (!bh) {
		/* XXX merge with other printk? */
                printk("TSTORE: unable to read superblock\n");
		goto outnobh;
	}

	if (!bh->b_data) {
		printk(KERN_ERR "TSTORE: no data in bh????\n");
		goto out;
	}

	if (memcmp(bh->b_data,"TSTORE5",8)) {
		if (!silent)
			printk ("VFS: Can't find a tstore filesystem.\n");
		goto out;
	}

	/* Find the start of the fs */
	ts_sb->numblocks = GET_BNUM(bh->b_data+8);
	ts_sb->lastentry = GET_BNUM(bh->b_data+13);

	brelse(bh);

	s->s_magic = TSTORE_MAGIC;

	s->s_flags |= MS_RDONLY;

	s->s_op	= &tstore_ops;
	s->s_root = d_alloc_root(iget(s, ts_sb->numblocks));

	if (!s->s_root)
		goto outnobh;

	return 0;

out:
	brelse(bh);
outnobh:
	return -EINVAL; /* FIXME: Better error code */
}

/* That's simple too. */

static int
tstore_statfs(struct super_block *sb, struct kstatfs *buf)
{
	struct tstore_super *ts_sb=sb->s_fs_info;
	ENTER();
	buf->f_type = TSTORE_MAGIC;
	buf->f_bsize = 512;
	buf->f_bfree = buf->f_bavail = buf->f_ffree = 0;
	buf->f_blocks = ts_sb->numblocks;
	buf->f_namelen = 494;
	return 0;
}

static sector_t tstore_block_number(struct super_block *sb,
				    long ino,
				    sector_t block)
{
	int levels;
	struct buffer_head *bh;
	uint8_t *buf;
	uint64_t block_number,file_size;
	
	bh=sb_bread(sb,ino);
	if (!bh) {
		printk("TSTORE: Read error on file header!\n");
		return -EIO;
	}

	buf=bh->b_data;

	block_number=GET_BNUM(buf+5);
	file_size=GET_SIZE(buf+10);
	brelse(bh);

	/* Cannot overflow because file_size does not use all 64 bits */
	if (block >= (file_size+511)/512) {
		printk(KERN_ERR "TSTORE: get_block outside file"
		       " - block %llu size %llu\n",
		       block,file_size);
		return 0; /* Looks like ext2 and minix are going to return 0 in this case */
	}
	levels=filelevels(file_size);

	while ((--levels)>=0) {
		int idx=block,x;
		for (x=0;x<levels;++x) idx/=102;
		idx%=102;

		if (block_number & (3ULL<<38)) {
			block_number = (block_number - (1ULL<<38) + idx);
		} else if (block_number) {
			bh=sb_bread(sb,block_number);
			if (!bh) {
				printk("TSTORE: Read error on indirect block?\n");
				return -EIO;
			}
			block_number=GET_BNUM(bh->b_data+5*idx);
			brelse(bh);
		}
	}

	return block_number;
}

static int tstore_get_block(struct inode *inode, sector_t block,
			    struct buffer_head *bh_result, int create)
{
	sector_t trans_block;
	/* create should never be set. And if it
	 * happens anyway I'd rather crash right
	 * here than corrupt my storage.
	 */
	BUG_ON(create);
	if (create) panic("TSTORE: This is not possible\n");
	BUG_ON(block<0);

	trans_block=tstore_block_number(inode->i_sb,inode->i_ino,block);
	if (trans_block<1) return trans_block;

	if (trans_block & (3ULL<<38)) {
		printk(KERN_ERR "TSTORE: Final sequential - no longer allowed\n");
		return -EIO;
	}

	map_bh(bh_result,inode->i_sb,trans_block);
        return 0;
}

static int tstore_readpage(struct file *file, struct page *page)
{
  return block_read_full_page(page,tstore_get_block);
}
static sector_t tstore_bmap(struct address_space *mapping, sector_t block)
{
  return generic_block_bmap(mapping,block,tstore_get_block);
}


/*
 * Lookup the data. This needs to be replaced.
 */
static struct dentry * tstore_lookup(struct inode *dir, struct dentry *dentry, struct nameidata *nd)
{
	int i;
	struct super_block *sb=dir->i_sb;
	struct tstore_super *ts_sb=sb->s_fs_info;
	for (i=ts_sb->lastentry;i;) {
		struct buffer_head *bh;
		uint8_t *buf;

		bh=sb_bread(sb, i);
		if (!bh) return ERR_PTR(-EIO);

		buf=bh->b_data;
		if (!strcmp(dentry->d_name.name,buf+17)) {
			struct inode *inode;
			brelse(bh);
			inode=iget(sb,i);
			if (!inode)
				return ERR_PTR(-EACCES);
			d_add(dentry,inode);
			return NULL;
		}
		i=GET_BNUM(buf);
		brelse(bh);
	}
        d_add(dentry, NULL);
        return NULL;
}

static int
tstore_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
	struct dentry *dentry = filp->f_dentry;
	struct super_block *sb = dentry->d_inode->i_sb;
	struct tstore_super *ts_sb=sb->s_fs_info;
	ino_t ino;
	ENTER();

	switch (filp->f_pos) {
#if 1
		case 0:
			ino = dentry->d_inode->i_ino;
			if (filldir(dirent, ".", 1, filp->f_pos, ino, DT_DIR) < 0)
				break;
			filp->f_pos=1;
			/* fallthrough */
		case 1:
			spin_lock(&dcache_lock);
			ino = dentry->d_parent->d_inode->i_ino;
			spin_unlock(&dcache_lock);
			if (filldir(dirent, "..", 2, filp->f_pos, ino, DT_DIR) < 0)
				break;
			filp->f_pos=ts_sb->lastentry+2;
			/* fallthrough */
#else
		case 0:
			filp->f_pos=ts_sb->lastentry+2;
			/* fallthrough */
#endif
		default:
			while(filp->f_pos > 2) {
				struct buffer_head *bh;
				uint8_t *buf;

				bh = sb_bread(sb,filp->f_pos-2);
				if (!bh) {
					printk("TSTORE: Read error on directory entry!\n");
					return -EIO;
				}

				buf=bh->b_data;

				if (filldir(dirent, buf+17, strlen(buf+17), filp->f_pos, filp->f_pos-2, DT_REG)) {
					brelse(bh);
					return 0;
				}

				filp->f_pos=GET_BNUM(buf)+2;

				brelse(bh);
			}
	}
	return 0;
}

/* Mapping from our types to the kernel */

static struct address_space_operations tstore_aops = {
	readpage:	tstore_readpage,
	bmap:		tstore_bmap,
};

static struct file_operations tstore_dir_operations = {
	read:		generic_read_dir,
	readdir:	tstore_readdir,
};

static struct inode_operations tstore_dir_inode_operations = {
	lookup:		tstore_lookup,
};

static void tstore_read_inode(struct inode *inode)
{
	struct tstore_super *ts_sb=inode->i_sb->s_fs_info;
	struct timespec ts={tv_sec:0,tv_nsec:0};
	int ino=inode->i_ino;
	ENTER();

	inode->i_uid = 0;
	inode->i_gid = 0;
	inode->i_nlink = 1;
	inode->i_size = 0;
	inode->i_blocks = inode->i_blksize = 0;
	inode->i_mode = S_IRUGO|S_IXUGO;
	inode->i_mapping->a_ops = &tstore_aops;
	inode->i_data.a_ops = &tstore_aops;

	if (ino < ts_sb->numblocks) {
		struct buffer_head *bh;
		uint8_t *buf;
		/* file */

		bh=sb_bread(inode->i_sb,ino);
		if (!bh) {
			printk("TSTORE: Read error on inode!\n");
		} else {
			buf=bh->b_data;
			inode->i_mode |= S_IFREG;
			inode->i_fop = &generic_ro_fops;
			inode->i_size = GET_SIZE(buf+10);
			ts.tv_sec = GET_BNUM(buf+507);
		}
	} else {
		/* root directory */
		inode->i_mode |= S_IFDIR;
		inode->i_size = 42;
		inode->i_nlink = 2;
		inode->i_op = &tstore_dir_inode_operations;
		inode->i_fop = &tstore_dir_operations;
		ts = CURRENT_TIME;
	}

        inode->i_mtime = inode->i_atime = inode->i_ctime = ts;
}

static struct super_block *tstore_get_sb(struct file_system_type *fs_type,
                                        int flags, const char *dev_name, void *data)
{
	return get_sb_bdev(fs_type, flags, dev_name, data, tstore_fill_super);
}

static struct super_operations tstore_ops = {
	read_inode:	tstore_read_inode,
	statfs:		tstore_statfs,
};

static struct file_system_type tstore_fs_type = {
	.owner		= THIS_MODULE,
	.name		= "tstore",
	.get_sb		= tstore_get_sb,
	.kill_sb	= kill_block_super,
	.fs_flags	= FS_REQUIRES_DEV,
};

static int __init init_tstore_fs(void)
{
	return register_filesystem(&tstore_fs_type);
}

static void __exit exit_tstore_fs(void)
{
	unregister_filesystem(&tstore_fs_type);
}

module_init(init_tstore_fs)
module_exit(exit_tstore_fs)
MODULE_LICENSE("GPL");
